using System;
using System.Net.Sockets;
using System.Runtime.Serialization.Formatters.Binary;
using System.Collections;
using System.Windows.Forms;
using Shared;
using System.Xml;


namespace Client
{
	/// <summary>
	/// The underlying code for the election client: this is where all the 
	/// work gets done. User actions through the user interface get resolved 
	/// into calls to this class, and it calls user interface (ClientDisplay) 
	/// methods to display results. The client communicates with the server 
	/// over a single connection. 
	/// </summary>
	public class Client
	{
		private String server;
		private short port;
		private TcpClient client;
		private BinaryFormatter bf;

		/// <summary>
		/// Constructor. Parses XML file for the server's IP addr and port
		/// </summary>
		/// <param name="file">Name of the configuration file</param>
		public Client(string file)
		{
			//parse an XML file 
			try
			{
				XmlTextReader reader = new XmlTextReader(file);
				while (reader.Read()) 
				{
					reader.Read();
					reader.Read();
					if((reader.Name == "config") && 
						(reader.NodeType == XmlNodeType.Element))
					{
						reader.Read();
						reader.Read();
						if((reader.Name == "port") && 
							(reader.NodeType == XmlNodeType.Element))
						{
							reader.Read();	
							this.port = Convert.ToInt16(reader.Value);
						}
						reader.Read();	reader.Read();	reader.Read();

						if((reader.Name == "server") && 
							(reader.NodeType == XmlNodeType.Element))
						{
							reader.Read();	
							this.server = reader.Value;
						}
					}
				}
			}
			catch(Exception)
			{
				MessageBox.Show("File not found!", "Error");
			}
		}

	

		/// <summary>
		/// Function called by GUI after the user enters the name of the 
		/// election. This is where the client makes TCP connection with 
		/// the server, and sends a request for user-specified election.
		/// </summary>
		/// <param name="election">Name of the election</param>
		/// <param name="display">Display Manager is passed along to display 
		/// the information</param>
		public void displayCandidates(string election, ClientDisplay display)
		{
			//create new message
			ArrayList list = new ArrayList();
			list.Add(election);
			Shared.Message request = new Shared.Message(
				Shared.Message.QUERY_CANDIDATES, list);
			Shared.Message reply;

			try
			{
				client = new TcpClient(server, port);
			}
			catch(Exception)
			{	throw new ServerUnreachableException
					("The server is unreachable");}
			bf = new BinaryFormatter();
    
			try 
			{	//request the list of candidates
				bf.Serialize(client.GetStream(), request);
				//wait for response
				reply = (Shared.Message)bf.Deserialize(client.GetStream());
			}
			catch (Exception e) 
			{
				throw new ClientException("Invalid Message received from "+
					"the server. Stack trace: "+ e.StackTrace);
			}
			
			//check the message context
			switch (reply.type) 
			{
					//expected reply
				case Shared.Message.LIST_CANDIDATES:
					if (reply.items == null)
						throw new MalformedMessageException("Error: "+
							"malformed message.");

					//call the methods in ClientDisplay object to
					//display available candidates
					IEnumerator enumerator = reply.items.GetEnumerator();
					while (enumerator.MoveNext()) 
					{
						String candidate = (String) enumerator.Current;
						display.putCandidate(candidate);
					}
					break;

				case Shared.Message.BAD_ELECTION:
					throw new DisplayException("Error: election closed or "+
						"nonexistent.");
					
				default:
					throw (new MalformedMessageException("Error: invalid "+
						"message type."));
			}
		}

		/// <summary>
		/// Function called by the GUI after the user selects one of the 
		/// candidates. Creates a new message with the election name and
		/// candidate selection, and sends the vote request to the server
		/// </summary>
		/// <param name="election">Name of election</param>
		/// <param name="candidate">Candidate selection</param>
		/// <param name="display">Display Manager</param>
		public void vote(string election, string candidate, 
			VoteDisplay display)
		{
			//create the vote request message
			ArrayList list = new ArrayList();
			list.Add(election);
			list.Add(candidate);
			Shared.Message request = new Shared.Message
				(Shared.Message.SEND_VOTE, list);
			Shared.Message reply;
			
			try 
			{
				//send the request
				bf.Serialize(client.GetStream(), request);
				//wait for the response and deserialize it
				reply = (Shared.Message)bf.Deserialize(client.GetStream());
			}
			catch (Exception) 
			{
				throw new MalformedMessageException();
			}
			//display the confirmation or throw the Exception that the vote 
			//was not casted
			switch (reply.type) 
			{
					//confirm that the vote was accepted
				case Shared.Message.ACK_VOTE:
					MessageBox.Show("Vote Tallied!");
					display.Dispose();
					break;
					//error handling
				case Shared.Message.BAD_ELECTION:
					throw new DisplayException("Error: election closed or "+
						"nonexistent.");
				case Shared.Message.BAD_CANDIDATE:
					throw new DisplayException("Error: nonexistent "+
						"candidate.");
				default:
					throw new MalformedMessageException("Error: invalid "+
						"message type.");
			}
		}
	}	
}
