using System;
using System.Collections;
using System.Runtime.CompilerServices;




namespace Server
{
	/// <summary>
	/// Implementation for a single election, which has a string name and 
	/// a list of candidates and count of votes for each candidate. Public 
	/// methods called by the election server manager (results and 
	/// exceptions are sent to the display), and methods called by Worker 
	/// threads. 
	/// <br/><br/>
	/// All methods declared synchronized to ensure the correct concurrency 
	/// control between multiple threads.
	/// </summary>
	public class Election
	{
		private const int CLOSED = 0;
		private const int OPEN = 1;
		
		/// <summary>
		/// Name of the election
		/// </summary>
		public string name;
		private ArrayList candidates;
		private Hashtable votes;
		private int status;

		/// <summary>
		/// Constructor.
		/// </summary>
		/// <remarks>All the elections are created <c>CLOSED</c></remarks>
		/// <param name="name">The name of the election</param>
		public Election(string name)
		{
			this.name = name;
			candidates = new ArrayList();
			votes = new Hashtable();
			status = CLOSED;
		}

		/// <summary>
		/// Adds a candidate to the election. <c>This</c> object is locked 
		/// for any other threads when this function is called.
		/// </summary>
		/// <param name="name">The name of the candidate</param>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void addCandidate(String name) 
		{
			candidates.Add(name);
			votes.Add(name, 0);
		}

		
		/// <summary>
		/// Returns the list of the candidates. <c>This</c> object is locked 
		/// for any other threads when this function is called.
		/// </summary>
		/// <returns>Candaidates' list</returns>
		/// <exception cref="ElectionNotOpenException">Raises the exception
		/// when the election is closed.
		/// </exception>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public ArrayList getCandidates() 
		{
			if (status == OPEN)
				return candidates;
			else
				throw new ElectionNotOpenException();
	}

		
		/// <summary>
		/// Opens the election. Raises <c>DisplayException</c> if the 
		/// election is already open. Resets the number of votes for each 
		/// candidate to 0.
		/// <c>This</c> object is locked 
		/// for any other threads when this function is called.
		/// </summary>
		/// <exception cref="DisplayException">Raises the exception when 
		/// the election is already open</exception>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void open() 
		{
			if (status == OPEN)
				throw new DisplayException("Error: election is already open.");
			
			IEnumerator enumerator = votes.Keys.GetEnumerator();
			votes = new Hashtable();
			while (enumerator.MoveNext()) 
			{
				votes[(string) enumerator.Current] = 0;
			}
    		status = OPEN;    
		}

		/// <summary>
		/// Closes the election.
		/// </summary>
		/// <param name="display">Display manager used to format the 
		/// <c>DisplayResults</c></param>
		/// <exception cref="DisplayException">Raised when election is 
		/// already closed</exception>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void close(ServerDisplay display) 
		{
			if (status == CLOSED)
				throw new DisplayException("Error: election is alreay closed."); 
			display.putTitle("Results of "+name);
		
			IEnumerator enumerator = votes.Keys.GetEnumerator();
			int maxvotes = -1;
			String candidate;
			String winner = null;

			while (enumerator.MoveNext()) 
			{
				candidate = (string) enumerator.Current;
				int nvotes = (int) votes[candidate];
				if (nvotes!=0)
					if (nvotes > maxvotes) 
					{
						winner = candidate;
						maxvotes = nvotes;
					}
					else if (nvotes == maxvotes)
					{
						//records multiple names for the winner if number of 
						//votes are equal
						winner = winner + " and " + candidate;
						maxvotes = nvotes;
					}
			}
			
			display.putWinner(winner, maxvotes);

			enumerator = votes.Keys.GetEnumerator();

			while (enumerator.MoveNext()) 
			{
				display.putCandidate((String) enumerator.Current, 
					(int) votes[(String) enumerator.Current]);
			}

			status = CLOSED;
		}

		/// <summary>
		/// Sets the <c>ServerDisplay</c> to display the results of the 
		/// election
		/// </summary>
		/// <exception cref="DisplayException">raises when the election is 
		/// <c>OPEN</c></exception>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void view(ServerDisplay display) 
		{
			if (status == OPEN)
				throw new DisplayException("Error: can only view a closed election.");
							
			display.putTitle("Votes in "+name);

			IEnumerator enumerator = votes.Keys.GetEnumerator();

			while (enumerator.MoveNext()) 
			{
				display.putCandidate((String) enumerator.Current, (int) votes[(String) enumerator.Current]);
			}
		}

		/// <summary>
		/// Adds an additional vote for the candidate
		/// </summary>
		/// <param name="candidate">The name of the candidate</param>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public void vote(String candidate) 
		{
			if (status == CLOSED)
				throw new ElectionNotOpenException();  
			if (votes[candidate] == null)
				throw new UndefinedCandidateException();  
			else
			{
				votes[candidate] = ((int) votes[candidate]) + 1;
			}
		}
		
		/// <summary>
		/// Checks if the election is open
		/// </summary>
		/// <returns>Whether the election is open</returns>
		[MethodImpl(MethodImplOptions.Synchronized)]
		public bool isOpen()
		{
			return status == OPEN;
		}
	}	
}
