using System;
using System.Collections;
using System.Xml;
using System.Net;


namespace Server
{
	/// <summary>
	/// The actual code for operating on elections: maintains the collection 
	/// of elections and has methods which Worker threads and the election 
	/// server manager can call. This class is effectively immutable: all 
	/// operations are passed through a mapping of election names to 
	/// Election objects, and the mapping does not change once the server 
	/// is up and running. Concurrency control is on individual elections, 
	/// since we don't need concurrency control on the server itself. 
	/// <br/><br/>
	/// Operations called by the election server manager have a display 
	/// argument, which is how they display results (i.e. no return values); 
	/// they may also throw a DisplayException to display an error. 
	/// Operations called by worker threads throw exceptions to indicate 
	/// errors, and return results normally. 
	/// </summary>

	public class Server
	{
		private short port;
		private IPAddress localAddr = IPAddress.Parse("127.0.0.1");
		private Hashtable elections;
		private const int DEFAULT_PORT=6789;

		/// <summary>
		/// Constructor. 
		/// </summary>
		public Server()
		{
			elections = new Hashtable();
		}

		/// <summary>
		/// Port accessor
		/// </summary>
		public short Port
		{
			get { return port; }
			set { this.port = value; }
		}

		/// <summary>
		/// Local address accessor
		/// </summary>
		public IPAddress LocalAddr
		{
			get { return localAddr; }
		}
 
		/// <summary>
		/// Reads the contents of an XML file and parses the elections and 
		/// the candidates.
		/// </summary>
		/// <param name="file">File name</param>
		/// <exception cref="Exception">Thrown if the file is not well formatted.</exception>
		public void setup(String file)
		{
			try
			{
				XmlTextReader reader = new XmlTextReader(file);
				while (reader.Read()) 
				{
					reader.Read();
					reader.Read();
					if((reader.Name == "config") && (reader.NodeType == XmlNodeType.Element))
					{
						reader.Read();
						reader.Read();
						if((reader.Name == "port") && (reader.NodeType == XmlNodeType.Element))
						{
							reader.Read();	
							this.port = Convert.ToInt16(reader.Value);
						}
						reader.Read();
						reader.Read();
						reader.Read();
						reader.Read();
						while(reader.Read())
						{
							if((reader.Name == "election") && (reader.NodeType == XmlNodeType.Element))
							{
								Election election;
								election = new Election(reader.GetAttribute("name"));
								while(reader.Name != "candidate")
									reader.Read();
								while(reader.NodeType != XmlNodeType.EndElement)
								{
									while(reader.NodeType != XmlNodeType.Text)
										reader.Read();
									election.addCandidate(reader.Value);
									while(reader.NodeType != XmlNodeType.EndElement)
										reader.Read();
									reader.Read();
									reader.Read();	
								}
								elections.Add(election.name, election);
							}
						}
					}
				}	
			}
			catch(XmlException e) 
			{
				throw new VotingAppException("Invalid XML format. Stack trace: "+ e.StackTrace);
			}
		}

		/// <summary>
		/// Opens the election by calling on the method on election object
		/// </summary>
		/// <param name="name">Name of the election to be opened</param>
		/// <exception cref="UndefinedElectionException">Thrown when the 
		/// wrong election name is passed</exception>

		public void openElection(String name) 
		{
			if (elections[name]!=null)
				((Election)elections[name]).open();
			else
				throw new UndefinedElectionException();
		}

		/// <summary>
		/// Closes the election by calling close() method in election.
		/// Passes Display Manager so that election can pass the 
		/// information to be displayed on the screen.
		/// </summary>
		/// <param name="name">Name of the election</param>
		/// <param name="display">Display Manager</param>
		/// <exception cref="UndefinedElectionException">Thrown when the 
		/// wrong election name is passed</exception>
		public void closeElection(String name, ServerDisplay display) 
		{
			if (elections[name]!=null)
				((Election)elections[name]).close(display);
			else
				throw new UndefinedElectionException();
		}

		/// <summary>
		/// Calls view method on election to display the election results.
		/// </summary>
		/// <param name="name">Name of the election</param>
		/// <param name="display">Display Manager</param>
		/// <exception cref="UndefinedElectionException">Thrown if the wrong 
		/// election name is passed</exception>
		public void viewElection(String name, ServerDisplay display) 
		{
			if (elections[name]!=null)
				((Election)elections[name]).view(display);
			else
				throw new UndefinedElectionException("Warning: Election doesn't exist");
		}

		/// <summary>
		/// Prepares Display Manager to display the results of the elections
		/// </summary>
		/// <param name="display">Display Manager</param>
		public void displayElections(ServerDisplay display) 
		{
			IEnumerator enumerator = elections.Keys.GetEnumerator();
			while (enumerator.MoveNext()) 
			{
				display.putElection((string) enumerator.Current);
			}
		}

		/// <summary>
		/// Gets the list of elections
		/// </summary>
		/// <returns>List of elections</returns>
		public ArrayList getElections()
		{
			ArrayList list = new ArrayList();
   			IEnumerator enumerator = elections.Keys.GetEnumerator();
			String election;

			while (enumerator.MoveNext()) 
			{
				election = (String) enumerator.Current;
				if (((Election) elections[election]).isOpen())
					list.Add(election);
			}
    		return list;
		}


		/// <summary>
		/// Gets the candidate list for the specified election
		/// </summary>
		/// <param name="poll">Name of the election</param>
		/// <returns>Candidate list</returns>
		public ArrayList getCandidates(String poll)
		{
			Election election;
			if (elections[poll]==null)
				throw new UndefinedElectionException();
			else
				election = (Election) elections[poll];
			return election.getCandidates();
		}

		/// <summary>
		/// Records the vote for the specified candidate in the election
		/// </summary>
		/// <param name="poll">Election Name</param>
		/// <param name="candidate">Candidate's name</param>
		public void vote(string poll, string candidate)
		{
			Election election;
	    
			if(elections[poll] == null)
				throw new UndefinedElectionException();
			election = (Election)elections[poll];
			election.vote(candidate);
		}   
	}  
}
