using System;
using System.Windows.Forms;
using System.Net.Sockets;
using System.Threading;
using System.IO;

namespace Server
{
	/// <summary>
	/// Startup code for the election server. The server is in two "parts": 
	/// the GUI part for handling administrator's requests, and the network 
	/// part for handling clients' communication. Both the network and the 
	/// GUI are hooked up to the actual server logic through the Server 
	/// object, which we create here. When the GUI wants something done, 
	/// it calls the server logic and passes itself as an argument, 
	/// so that the server can display the results. The server reads in the 
	/// contents of config.xml and configures the port number listed in the 
	/// port tag, and sets up the elections.
	/// </summary>
	public class ServerMain 
	{
		/// <summary>
		/// Constructor
		/// </summary>
		public ServerMain()
		{}

		[STAThread]
		static void Main() 
		{
			try
			{
				Server server = null;
				ServerDisplay manager = null;
				TcpListener listener = null;
				Network network;
				
				server = new Server();
				//loads the election information and port information
				try
				{
					server.setup("config.xml");
				}
				catch(IOException e)
				{
					MessageBox.Show("File not found!", "Error");
					throw new VotingAppException("File was not found. Stack trace: " + e.StackTrace);
				}
				//Initiliazes the Display Manager
				manager = new ServerDisplay(server);
				manager.start();
				
				//Starts a listener on the server's default port
				listener = new TcpListener(server.LocalAddr, server.Port);
				listener.Start();

				//Instantiates a Network thread to respond to the clients
				network = new Network(server, listener);
				Thread networkThread = new Thread(new ThreadStart(network.run));
				networkThread.Start();  
				
				//Starts the Display manager
				Application.Run(manager);

				//terminates the network thread
				networkThread.Abort();
				
			}
			catch(VotingAppException e)
			{
				MessageBox.Show(e.Message + ": " + e.StackTrace, "Fatal Voting Error");
				Application.Exit();
			}
		}
	}
}
