using System;
using System.Windows.Forms;
using System.Net.Sockets;
using System.IO;
using System.Collections;
using System.Runtime.Serialization.Formatters.Binary;
using System.Runtime.Serialization;
using Shared;
using System.Threading;

namespace Server
{
	/// <summary>
	/// Server processing for an individual client request. When the voter 
	/// supplies an election name, the client code makes a TCP connection 
	/// to the server. The server instantiates a Worker thread that 
	/// processes client's request.
	/// <br/><br/>
	/// The worker thread expects to get a request for the candidates in an
	/// election, for which it returns the list of candidates. Then it expects 
	/// a vote for the election, which it processes and then acknowledges. 
	/// Any deviation from the expected message format, or invalid message 
	/// contents (closed election, etc) results in an error reply being sent. 
	/// <br/><br/>
	/// All the operations on elections are done by calling methods on the 
	/// <code cref="Election">Election</code> member variable, which is set at thread initialisation time. 
	/// The server handles concurrency control between multiple workers at 
	/// the level of individual elections. 
	/// </summary>
	public class Worker
	{
		Server server;
		Socket socket;

		/// <summary>
		/// Objected wrapped around the socket to communicate with the server
		/// </summary>
		NetworkStream stm;
		
		/// <summary>
		/// Constructor. Initializes the server, socket, and creates a 
		/// NetworkStream to communicate with the client.
		/// </summary>
		/// <param name="server">Contains the pointer to the server</param>
		/// <param name="socket">Contains the pointer to the socket</param>
		public Worker(Server server, Socket socket)
		{
			this.server = server;
			this.socket = socket;
			this.stm = new NetworkStream ( socket );
		}
		
		/// <summary>
		/// Method called by Network object to start processing client's 
		/// request and record the vote via initialized socket.
		/// </summary>
		/// <exception cref="MalformedMessageException">Exception thrown 
		/// if the client's message is not of a specific format
		/// </exception>
		public void run()
		{
			try 
			{
				processCandidatesRequest();
				processVoteRequest();
				socket.Close();
			}
			catch (MalformedMessageException e) 
			{
				throw new VotingAppException("Invalid Message Format. Stack trace: " 
					+ e.StackTrace);
			}
			catch (IOException e) 
			{
				throw new VotingAppException("IO Error. Stack trace: " + e.StackTrace);
			}
			catch (ThreadAbortException) 
			{}
			
		}

		/// <summary>
		/// Process the message with <c>QUERY_CANDIDATES</c> request type, and
		/// name of election, checks if this election exists, and responds 
		/// with the list of candidates. If the election is not defined or 
		/// closed, <c>BAD_ELECTION</c> message is sent back to the client.
		/// </summary>
		/// <remarks>Both client and the server share Message assmebly to allow 
		/// <c>Serialization</c> of the objects passed between them</remarks>
		private void processCandidatesRequest() 
		{
			Shared.Message reply = null;
			string election = null;
			BinaryFormatter bf = new BinaryFormatter();
			Shared.Message request;
			try
			{
				//receive the message
				request = (Shared.Message)bf.Deserialize(stm);
			}
			catch(Exception){return;}
			try
			{
				//check for the message type
				if (request.type != Shared.Message.QUERY_CANDIDATES)
					throw new MalformedMessageException("Error: not a candidates query.");
				//look up the candidate list for this election		
				election = (String) request.items[0];
				ArrayList candidates = server.getCandidates(election);
				//create a reply message
				reply = new Shared.Message(Shared.Message.LIST_CANDIDATES, candidates);
			}
			catch (ElectionNotOpenException) 
			{
				reply = new Shared.Message(Shared.Message.BAD_ELECTION);
			}
			catch (UndefinedElectionException) 
			{
				reply = new Shared.Message(Shared.Message.BAD_ELECTION);	
			}
			
			try 
			{	//send it
				bf.Serialize(stm, reply);
			}
			catch (Exception) {return;}
		}

		/// <summary>
		/// Process the message with <c>SEND_VOTE</c> request type, 
		/// name of election, and the vote. Casts the vote and sends the 
		/// <c>ACK_VOTE</c> message back to the client. If the election is 
		/// closed, doesn't exist, or the candidate is not part of the list,
		/// an error message is sent back to the client
		/// </summary>
		private void processVoteRequest()
		{
			Shared.Message reply = null;
			String election = null;
			String candidate = null;
			BinaryFormatter bf = new BinaryFormatter();
			Shared.Message request;
			try
			{
				//receive the message
				request = (Shared.Message)bf.Deserialize(stm);
			}
			catch(Exception) { return; }
			try 
			{	
				//check the type of the message
				if (request.type != Shared.Message.SEND_VOTE)
					throw new MalformedMessageException("Error: not a vote.");
				if (request.items == null)
					throw new MalformedMessageException("Error: malformed message.");

				election = (String) request.items[0];
				candidate = (String) request.items[1];
				
				//cast the vote
				server.vote(election, candidate);
				//create the reply message
				reply = new Shared.Message(Shared.Message.ACK_VOTE);
			}
			catch (ElectionNotOpenException) 
			{
				reply = new Shared.Message(Shared.Message.BAD_ELECTION);
			}
			catch (UndefinedElectionException) 
			{
				reply = new Shared.Message(Shared.Message.BAD_ELECTION);
			}
			catch (UndefinedCandidateException) 
			{
				reply = new Shared.Message(Shared.Message.BAD_CANDIDATE);
			}
						
			try 
			{
				//send the message back
				bf.Serialize(stm, reply);
			}
			catch (Exception){return;}
		}

	}

}
